"""App Engine Cron jobs"""

from typing import Tuple

from flask import Blueprint, abort, request

from ibkr_report.definitions import BUCKET_ID
from ibkr_report.exchangerates import ExchangeRates

bp = Blueprint("cron", __name__)


@bp.route("/cron", methods=["GET"])
def cron() -> Tuple[str, int]:
    """Cron function that fetches the latest exchange rates if necessary."""
    if request.headers.get("X-Appengine-Cron") is None:
        abort(403)
    if BUCKET_ID:
        try:
            _ = ExchangeRates()
        except ValueError as err:
            return str(err), 500
        return "Done!", 200
    return "BUCKET_ID missing!", 500



"""Extracts details about trades."""

import logging
from decimal import Decimal
from typing import Tuple

from ibkr_report.definitions import (
    AssetCategory,
    Field,
    ReportOptions,
    RowData,
    TradeDetails,
)
from ibkr_report.exchangerates import ExchangeRates
from ibkr_report.tools import (
    add_years,
    date_without_time,
    decimal_cleanup,
    get_date,
)

log = logging.getLogger(__name__)


class Trade:
    """Trade which might be related to several ClosedLot rows."""

    fee: Decimal = Decimal(0)
    closed_quantity: Decimal = Decimal(0)
    total_selling_price: Decimal = Decimal(0)
    data: RowData
    options: ReportOptions
    rates: ExchangeRates

    def __init__(
        self, items: Tuple[str, ...], options: ReportOptions, rates: ExchangeRates
    ) -> None:
        """Initializes the Trade and calculates the total selling price from it."""
        self.options = options
        self.rates = rates
        self.data = self._row_data(items)

        fee = decimal_cleanup(items[Field.COMMISSION_AND_FEES + self.options.offset])
        self.fee = fee / self.data.rate

        # Sold stocks have a negative value in the "Quantity" column
        if self.data.quantity < Decimal(0):
            proceeds = decimal_cleanup(items[Field.PROCEEDS + self.options.offset])
            self.total_selling_price = proceeds / self.data.rate
        log.debug(
            'Trade: "%s" "%s" %.2f',
            self.data.date_str,
            self.data.symbol,
            self.data.quantity,
        )

    def details_from_closed_lot(self, items: Tuple[str, ...]) -> TradeDetails:
        """Calculates the realized gains or losses from the ClosedLot related to the Trade."""
        lot_data = self._row_data(items)
        self._validate_lot(lot_data)

        sell_date = date_without_time(self.data.date_str)
        unit_sell_price = self.data.price_per_share
        buy_date = date_without_time(lot_data.date_str)
        unit_buy_price = lot_data.price_per_share

        # Swap if closing a short position
        if lot_data.quantity < Decimal(0):
            sell_date, buy_date = buy_date, sell_date
            unit_sell_price, unit_buy_price = unit_buy_price, unit_sell_price

        # One option represents 100 shares of the underlying stock
        multiplier = 100 if items[Field.ASSET_CATEGORY] == AssetCategory.OPTIONS else 1
        lot_sell_price = abs(lot_data.quantity) * unit_sell_price * multiplier
        lot_buy_price = abs(lot_data.quantity) * unit_buy_price * multiplier
        lot_fee = lot_data.quantity * self.fee / self.data.quantity
        realized = lot_sell_price - lot_buy_price - lot_fee
        if self.options.deemed_acquisition_cost:
            deemed_profit = self.deemed_profit(lot_sell_price, buy_date, sell_date)
            realized = min(realized, deemed_profit)

        log.info(
            "Symbol: %s, Quantity: %.2f, Buy date: %s, Sell date: %s, "
            "Selling price: %.2f, Gains/Losses: %.2f",
            lot_data.symbol,
            abs(lot_data.quantity),
            buy_date,
            sell_date,
            lot_sell_price,
            realized,
        )
        self.closed_quantity += lot_data.quantity
        if self.closed_quantity + self.data.quantity == Decimal(0):
            log.debug("All lots closed")
        return TradeDetails(
            symbol=lot_data.symbol,
            quantity=abs(lot_data.quantity),
            buy_date=buy_date,
            sell_date=sell_date,
            price=lot_sell_price,
            realized=realized,
        )

    def _row_data(self, items: Tuple[str, ...]) -> RowData:
        symbol = items[Field.SYMBOL + self.options.offset]
        date_str = items[Field.DATE_TIME + self.options.offset]
        rate = self.rates.get_rate(
            currency_from=self.options.report_currency,
            currency_to=items[Field.CURRENCY],
            date_str=date_str,
        )
        original_price_per_share = items[Field.TRANSACTION_PRICE + self.options.offset]
        price_per_share = decimal_cleanup(original_price_per_share) / rate
        quantity = decimal_cleanup(items[Field.QUANTITY + self.options.offset])
        return RowData(symbol, date_str, rate, price_per_share, quantity)

    def _validate_lot(self, lot_data: RowData) -> None:
        error_msg = ""

        if self.data.symbol != lot_data.symbol:
            error_msg = (
                f"Symbol mismatch! Date: {lot_data.date_str}, "
                f"Trade: {self.data.symbol}, ClosedLot: {lot_data.symbol}"
            )
        elif abs(self.data.quantity + lot_data.quantity) > abs(self.data.quantity):
            error_msg = (
                'Invalid data. "Trade" and "ClosedLot" quantities do not match. '
                f"Date: {lot_data.date_str}, Symbol: {lot_data.symbol}"
            )
        if error_msg:
            log.debug(lot_data)
            raise ValueError(error_msg)

    @staticmethod
    def deemed_profit(sell_price: Decimal, buy_date: str, sell_date: str) -> Decimal:
        """If you have owned the shares you sell for less than 10 years, the deemed
        acquisition cost is 20% of the selling price of the shares.
        If you have owned the shares you sell for at least 10 years, the deemed
        acquisition cost is 40% of the selling price of the shares.

        https://www.vero.fi/en/individuals/property/investments/selling-shares/
        """
        multiplier = Decimal(0.8)
        if get_date(buy_date) <= add_years(get_date(sell_date), -10):
            multiplier = Decimal(0.6)
        return multiplier * sell_price



""" IBKR Report Parser

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program. If not, see <https://www.gnu.org/licenses/>.
"""

from flask import Flask

from ibkr_report import cron, website
from ibkr_report.definitions import StorageType
from ibkr_report.exchangerates import ExchangeRates
from ibkr_report.report import Report
from ibkr_report.storage import get_storage

__all__ = ["ExchangeRates", "Report", "StorageType", "create_app", "get_storage"]


def create_app() -> Flask:
    """Create and configure an instance of the Flask application."""
    app = Flask(__name__)
    app.register_blueprint(cron.bp)
    app.register_blueprint(website.bp)

    return app



"""Tools and utility functions."""

from __future__ import annotations

import hashlib
import logging
import os
import re
from base64 import b64encode
from datetime import date, datetime
from decimal import Decimal
from typing import Any, Dict

from flask import current_app

from ibkr_report.definitions import _DEFAULT_LOGGING, DATE_STR_FORMATS, LOGGING_LEVEL

_cache: Dict = {}
_MAXCACHE = 10


class Cache:
    """Simple cache in memory"""

    @staticmethod
    def get(key: Any) -> Any:
        """Get value from cache, None if not found"""
        if key in _cache:
            return _cache[key]
        return None

    @staticmethod
    def set(key: Any, value: Any) -> None:
        """Set value into cache"""
        if key not in _cache and len(_cache) >= _MAXCACHE:
            try:
                del _cache[next(iter(_cache))]
            except (StopIteration, RuntimeError, KeyError):
                pass
        _cache[key] = value

    @staticmethod
    def clear() -> None:
        """Clear cache"""
        _cache.clear()


def get_date(date_str: str) -> date:
    """Converts a string formatted date to a date object."""
    for date_format in DATE_STR_FORMATS:
        try:
            return datetime.strptime(date_str, date_format).date()
        except ValueError:
            pass
    raise ValueError(f"Invalid date '{date_str}'")


def add_years(d_obj: date, years: int) -> date:
    """Return a date that's `years` years after the date (or datetime)
    object `d_obj`. Return the same calendar date (month and day) in the
    destination year, if it exists, otherwise use the previous day
    (thus changing February 29 to February 28).
    """
    try:
        return d_obj.replace(year=d_obj.year + years)
    except ValueError:
        return d_obj + (date(d_obj.year + years, 3, 1) - date(d_obj.year, 3, 1))


def date_without_time(date_str: str) -> str:
    """Strips away hours, minutes, and seconds from a string."""
    return re.sub(r"(\d\d\d\d-\d\d-\d\d),? ([0-9:]+)", r"\1", date_str)


def decimal_cleanup(number_str: str) -> Decimal:
    """Converts a string to a decimal while ignoring spaces and commas."""
    return Decimal(re.sub(r"[,\s]+", "", number_str))


def is_number(number_str: str) -> bool:
    """Checks if a string can be converted into a number"""
    try:
        float(number_str)
        return True
    except ValueError:
        return False


def calculate_sri_on_file(filename: str) -> str:
    """Calculate Subresource Integrity string."""
    hash_func = hashlib.sha384()
    update_hash(filename, hash_func)
    hash_base64 = b64encode(hash_func.digest()).decode()
    return f"sha384-{hash_base64}"


# TODO: mypy "BinaryIO" has no attribute "readinto"  # pylint: disable=fixme
#       https://github.com/python/mypy/issues/11917
# TODO: pylint "Module 'hashlib' has no '_Hash' member"  # pylint: disable=fixme
#       https://github.com/PyCQA/pylint/issues/5395
def update_hash(
    filename: str, hash_func: hashlib._Hash  # pylint: disable=no-member
) -> None:
    """Compute message digest from a file."""
    byte_array = bytearray(128 * 1024)
    memory_view = memoryview(byte_array)
    with open(filename, "rb", buffering=0) as file:
        for block in iter(lambda: file.readinto(memory_view), 0):  # type: ignore
            hash_func.update(memory_view[:block])


def set_logging() -> None:
    """Set logging level according to the ENV variable LOGGING_LEVEL."""
    level = logging.getLevelName(LOGGING_LEVEL.upper())
    if isinstance(level, int):
        current_app.logger.setLevel(level)
    else:
        current_app.logger.setLevel(logging.getLevelName(_DEFAULT_LOGGING))
    log_level = logging.getLevelName(current_app.logger.level)
    current_app.logger.debug(f"Logging level: {log_level}")


def sri(files: Dict[str, str]) -> Dict[str, str]:
    """Calculate Subresource Integrity for CSS and Javascript files.

    input: {'style.css': 'static/style.css', 'main.js': 'static/main.js', ...}
    output: {'style.css': 'sha384-...', 'main.js': 'sha384-...', ...}
    """
    cache_key = tuple(sorted(files.items()))
    sri_dict = Cache.get(cache_key)
    if not sri_dict:
        sri_dict = {}
        for key, file_path in files.items():
            sri_dict[key] = calculate_sri_on_file(file_path)
        Cache.set(cache_key, sri_dict)
    return sri_dict


def _sri() -> Dict[str, str]:
    return sri(
        {
            "main.css": os.path.join(
                current_app.root_path, "static", "css", "main.css"
            ),
            "main.js": os.path.join(current_app.root_path, "static", "js", "main.js"),
        }
    )



import os
import unittest
from decimal import Decimal
from unittest.mock import patch

from ibkr_report import Report
from ibkr_report.tools import Cache

THIS_PATH = os.path.abspath(os.getcwd())
TEST_URL = f"file://{THIS_PATH}/tests/test-data/eurofxref-hist.zip"


@patch("ibkr_report.exchangerates.EXCHANGE_RATES_URL", TEST_URL)
class ReportTest(unittest.TestCase):
    def setUp(self):
        Cache.clear()

    def test_without_deemed_cost(self):
        report = Report(use_deemed_acquisition_cost=False)
        with open("tests/test-data/data_deemed_acquisition_cost.csv", "rb") as file:
            report.add_trades(file)
        self.assertEqual(report.prices, Decimal("9982.0"))
        self.assertEqual(round(report.gains, 2), Decimal("6937.94"))
        self.assertEqual(report.losses, Decimal("0.00"))

    def test_report_currency_usd(self):
        report = Report(report_currency="USD", use_deemed_acquisition_cost=False)
        with open("tests/test-data/data_deemed_acquisition_cost.csv", "rb") as file:
            report.add_trades(file)
        self.assertEqual(round(report.prices, 2), Decimal("10957.24"))
        self.assertEqual(round(report.gains, 2), Decimal("6826.73"))
        self.assertEqual(round(report.losses, 2), Decimal("0.00"))

    def test_report_currency_eur_lowercase(self):
        report = Report(report_currency="eur")
        with open("tests/test-data/data_single_account.csv", "rb") as file:
            report.add_trades(file)
        self.assertEqual(round(report.prices, 2), Decimal("8518.52"))
        self.assertEqual(round(report.gains, 2), Decimal("5964.76"))
        self.assertEqual(round(report.losses, 2), Decimal("0.00"))


if __name__ == "__main__":
    unittest.main()



import os
import unittest
from pathlib import Path
from tempfile import mkdtemp
from unittest.mock import patch

from gcp_storage_emulator.server import create_server  # type: ignore
from moto import mock_s3  # type: ignore

from ibkr_report import StorageType, get_storage
from ibkr_report.tools import Cache

TEST_BUCKET = "test"


class StorageTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.test_data = {"2021-10-25": {"USD": "1.1603"}}
        os.environ["STORAGE_EMULATOR_HOST"] = "http://localhost:9023"
        cls._server = create_server(
            "localhost", 9023, in_memory=True, default_bucket=TEST_BUCKET
        )
        cls._server.start()

    @classmethod
    def tearDownClass(cls):
        cls._server.stop()

    def setUp(self):
        Cache.clear()
        self.storage_dir = Path(mkdtemp())

    def tearDown(self):
        for item in self.storage_dir.iterdir():
            item.unlink()
        self.storage_dir.rmdir()

    def test_non_existent_type(self):
        with self.assertRaises(NotImplementedError):
            get_storage("not-implemented")

    def test_disabled_storage(self):
        storage = get_storage()()
        storage.cache = False
        storage.save(self.test_data)
        self.assertEqual(storage.load(), {})

    @patch("ibkr_report.storage.BUCKET_ID", TEST_BUCKET)
    def test_google_cloud_storage_save_and_load(self):
        storage = get_storage(StorageType.GCP)()
        storage.cache = False
        storage.save(self.test_data)
        self.assertEqual(storage.load(), self.test_data)

    def test_gcp_load_not_existing(self):
        storage = get_storage(StorageType.GCP)(bucket_id=TEST_BUCKET)
        self.assertEqual(storage.load(), {})

    @mock_s3
    @patch("ibkr_report.storage.BUCKET_ID", TEST_BUCKET)
    def test_aws_s3_save_and_load(self):
        storage = get_storage(StorageType.AWS)()
        storage.cache = False
        storage.save(self.test_data)
        self.assertEqual(storage.load(), self.test_data)

    @mock_s3
    @patch("ibkr_report.storage.BUCKET_ID", TEST_BUCKET)
    def test_aws_s3_load_not_existing(self):
        storage = get_storage(StorageType.AWS)()
        self.assertEqual(storage.load(), {})

    def test_local_save_and_load(self):
        storage = get_storage(StorageType.LOCAL)(storage_dir=self.storage_dir)
        storage.cache = False
        storage.save(self.test_data)
        self.assertEqual(storage.load(), self.test_data)

    def test_local_load_not_existing(self):
        storage = get_storage(StorageType.LOCAL)(storage_dir=self.storage_dir)
        self.assertEqual(storage.load(), {})

    def test_local_custom_file_save_and_load(self):
        storage = get_storage(StorageType.LOCAL)(storage_dir=self.storage_dir)
        storage.save(self.test_data, "my-test-file")
        Cache.clear()
        self.assertEqual(storage.load("my-test-file"), self.test_data)
        self.assertEqual(storage.load("not-existing"), {})
        full_path = self.storage_dir.joinpath("my-test-file")
        self.assertTrue(full_path.is_file())

    def test_generating_filenames(self):
        storage = get_storage()()
        name1 = storage.get_filename("test1")
        name2 = storage.get_filename("test2")
        self.assertNotEqual(name1, name2)
        name_default = storage.get_filename()
        self.assertNotEqual(name1, name_default)
        self.assertNotEqual(name2, name_default)

    @patch("ibkr_report.storage.BUCKET_ID", TEST_BUCKET)
    def test_bucket_defined_but_type_not_defined(self):
        with self.assertRaises(ValueError):
            get_storage()


if __name__ == "__main__":
    unittest.main()



import json
import os
import unittest
from tempfile import NamedTemporaryFile
from unittest.mock import patch
from urllib.error import HTTPError

from gcp_storage_emulator.server import create_server  # type: ignore

from ibkr_report import create_app
from ibkr_report.definitions import FieldValue, _strtobool
from ibkr_report.tools import Cache, calculate_sri_on_file

TEST_BUCKET = "test"
THIS_PATH = os.path.abspath(os.getcwd())
TEST_URL = f"file://{THIS_PATH}/tests/test-data/eurofxref-hist.zip"
TEST_BROKEN_URL = f"file://{THIS_PATH}/tests/test-data/eurofxref-broken.csv"

# echo -n "" | openssl dgst -sha384 -binary | openssl base64 -A
EMPTY_FILE_SRI = (
    "sha384-OLBgp1GsljhM2TJ+sbHjaiH9txEUvgdDTAzHv2P24donTt6/529l+9Ua0vFImLlb"
)
TEST_STRING = "alert('Hello, world.');"
# echo -n "alert('Hello, world.');" | openssl dgst -sha384 -binary | openssl base64 -A
TEST_STRING_SRI = (
    "sha384-H8BRh8j48O9oYatfu5AZzq6A9RINhZO5H16dQZngK7T62em8MUt1FLm52t+eX6xO"
)


@patch("ibkr_report.exchangerates.EXCHANGE_RATES_URL", TEST_URL)
class SmokeTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        os.environ["STORAGE_EMULATOR_HOST"] = "http://localhost:9023"
        os.environ["GOOGLE_CLOUD_PROJECT"] = "<test>"
        cls._server = create_server(
            "localhost", 9023, in_memory=True, default_bucket=TEST_BUCKET
        )
        cls._server.start()

    @classmethod
    def tearDownClass(cls):
        cls._server.stop()

    def setUp(self):
        app = create_app()
        self.app = app.test_client()
        self.assertEqual(app.debug, False)

    def test_get_main_page(self):
        response = self.app.get("/")
        self.assertEqual(response.status_code, 200)

    @patch("ibkr_report.storage.BUCKET_ID", TEST_BUCKET)
    @patch("ibkr_report.cron.BUCKET_ID", TEST_BUCKET)
    @patch("ibkr_report.storage.STORAGE_TYPE", "gcp")
    def test_get_cron(self):
        response = self.app.get("/cron")
        self.assertEqual(response.status_code, 403)
        response = self.app.get("/cron", headers={"X-Appengine-Cron": "true"})
        self.assertEqual(response.data, b"Done!")
        self.assertEqual(response.status_code, 200)

    def test_get_cron_without_bucket(self):
        response = self.app.get("/cron", headers={"X-Appengine-Cron": "true"})
        self.assertEqual(response.status_code, 500)
        self.assertEqual(response.data, b"BUCKET_ID missing!")

    @patch("ibkr_report.cron.BUCKET_ID", TEST_BUCKET)
    @patch("ibkr_report.storage.BUCKET_ID", TEST_BUCKET)
    def test_get_cron_with_storage_disabled(self):
        response = self.app.get("/cron", headers={"X-Appengine-Cron": "true"})
        self.assertEqual(response.status_code, 500)
        msg = (
            b"[BUCKET_ID] set as 'test', but [STORAGE_TYPE] is set as 'disabled'."
            + b" With a bucket [STORAGE_TYPE] needs to be set as [AWS|GCP]."
        )
        self.assertEqual(response.data, msg)

    def test_post_single_account(self):
        data = {"file": open("tests/test-data/data_single_account.csv", "rb")}
        response = self.app.post("/result", data=data)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.headers.get("Content-Type"), "text/html; charset=utf-8"
        )

    def test_post_single_account_json(self):
        data = {"file": open("tests/test-data/data_single_account.csv", "rb")}
        response = self.app.post("/result?json", data=data)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers.get("Content-Type"), "application/json")
        data_json = json.loads(response.data)
        self.assertEqual(data_json["prices"], 8518.52)
        self.assertEqual(data_json["gains"], 5964.76)
        self.assertEqual(data_json["losses"], 0)
        self.assertIsInstance(data_json["details"], list)

    def test_post_multi_account_json(self):
        data = {"file": open("tests/test-data/data_multi_account.csv", "rb")}
        response = self.app.post("/result?json", data=data)
        self.assertEqual(response.status_code, 200)
        data_json = json.loads(response.data)
        self.assertEqual(data_json["prices"], 6873.02)
        self.assertEqual(data_json["gains"], 1064.02)
        self.assertEqual(data_json["losses"], 445.98)

    def test_post_data_in_future(self):
        data = {"file": open("tests/test-data/data_dates_in_future.csv", "rb")}
        response = self.app.post("/result", data=data)
        self.assertEqual(response.status_code, 400)

    def test_post_shorting_stocks(self):
        data = {"file": open("tests/test-data/data_shorting_stocks.csv", "rb")}
        response = self.app.post("/result?json", data=data)
        self.assertEqual(response.status_code, 200)
        data_json = json.loads(response.data)
        self.assertEqual(data_json["prices"], 6034.30)
        self.assertEqual(data_json["gains"], 2644.18)
        self.assertEqual(data_json["losses"], 0.00)

    def test_post_shorting_not_closed(self):
        data = {"file": open("tests/test-data/data_shorting_not_closed.csv", "rb")}
        response = self.app.post("/result?json", data=data)
        self.assertEqual(response.status_code, 200)
        data_json = json.loads(response.data)
        self.assertEqual(data_json["prices"], 6034.30)
        self.assertEqual(data_json["gains"], 0.00)
        self.assertEqual(data_json["losses"], 0.00)

    def test_post_deemed_acquisition_cost(self):
        data = {"file": open("tests/test-data/data_deemed_acquisition_cost.csv", "rb")}
        response = self.app.post("/result?json", data=data)
        self.assertEqual(response.status_code, 200)
        data_json = json.loads(response.data)
        self.assertEqual(data_json["prices"], 9982.0)
        self.assertEqual(data_json["gains"], 5989.2)
        self.assertEqual(data_json["losses"], 0.00)

    def test_post_closedlot_without_trade(self):
        data = {"file": open("tests/test-data/data_closedlot_without_trade.csv", "rb")}
        response = self.app.post("/result", data=data)
        self.assertEqual(response.status_code, 400)

    def test_post_numbers_within_quotes(self):
        data = {"file": open("tests/test-data/data_numbers_within_quotes.csv", "rb")}
        response = self.app.post("/result?json", data=data)
        self.assertEqual(response.status_code, 200)
        data_json = json.loads(response.data)
        self.assertEqual(data_json["prices"], 7195.37)
        self.assertEqual(data_json["gains"], 455.67)
        self.assertEqual(data_json["losses"], 0.00)

    def test_post_currency_rate_over_1000(self):
        data = {"file": open("tests/test-data/data_currency_rate_over_1000.csv", "rb")}
        response = self.app.post("/result?json", data=data)
        self.assertEqual(response.status_code, 200)
        data_json = json.loads(response.data)
        self.assertEqual(data_json["prices"], 7.2)
        self.assertEqual(data_json["gains"], 5.0)
        self.assertEqual(data_json["losses"], 0.00)

    @patch("ibkr_report.tools.LOGGING_LEVEL", "INVALID_LOGGING_LEVEL")
    def test_post_invalid_date_html(self):
        data = {"file": open("tests/test-data/data_invalid_date.csv", "rb")}
        response = self.app.post("/result", data=data)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.headers.get("Content-Type"), "text/html; charset=utf-8"
        )

    def test_post_invalid_data(self):
        data = {"file": open("tests/test-data/eurofxref-hist.zip", "rb")}
        response = self.app.post("/result", data=data)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.headers.get("Content-Type"), "text/html; charset=utf-8"
        )

    def test_post_invalid_date_json(self):
        data = {"file": open("tests/test-data/data_invalid_date.csv", "rb")}
        response = self.app.post("/result?json", data=data)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.headers.get("Content-Type"), "application/json")
        data_json = json.loads(response.data)
        self.assertEqual(
            data_json["error"], "400 Bad Request: Invalid date '20xx-08-23, 09:33:11'"
        )

    def test_post_symbol_mismatch(self):
        data = {"file": open("tests/test-data/data_symbol_mismatch.csv", "rb")}
        response = self.app.post("/result?json", data=data)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.headers.get("Content-Type"), "application/json")
        data_json = json.loads(response.data)
        self.assertEqual(
            data_json["error"],
            "400 Bad Request: Symbol mismatch! Date: 2021-03-15, Trade: UPST, ClosedLot: XXXX",
        )

    def test_post_wrong_quantities(self):
        data = {"file": open("tests/test-data/data_wrong_quantities.csv", "rb")}
        response = self.app.post("/result?json", data=data)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.headers.get("Content-Type"), "application/json")
        data_json = json.loads(response.data)
        self.assertEqual(
            data_json["error"],
            '400 Bad Request: Invalid data. "Trade" and "ClosedLot" quantities do not match. Date: 2021-03-15, Symbol: UPST',
        )

    def test_404_error_from_downloading(self):
        self._server.wipe()
        Cache.clear()
        with patch(
            "ibkr_report.exchangerates.urlopen",
            side_effect=HTTPError("", 404, "Test", {}, None),
        ):
            data = {"file": open("tests/test-data/data_single_account.csv", "rb")}
            response = self.app.post("/result?json", data=data)
            self.assertEqual(response.status_code, 400)
            data_json = json.loads(response.data)
            self.assertEqual(
                data_json["error"],
                "400 Bad Request: Maximum number of retries exceeded. Could not retrieve currency exchange rates.",
            )

    @patch("ibkr_report.tools._MAXCACHE", 0)
    @patch("ibkr_report.storage.BUCKET_ID", TEST_BUCKET)
    @patch("ibkr_report.storage.STORAGE_TYPE", "gcp")
    @patch("ibkr_report.cron.BUCKET_ID", TEST_BUCKET)
    def test_caching_filled_from_cron(self):
        self._server.wipe()
        Cache.clear()
        response = self.app.get("/cron", headers={"X-Appengine-Cron": "true"})
        self.assertEqual(response.status_code, 200)
        data = {"file": open("tests/test-data/data_single_account.csv", "rb")}
        response = self.app.post("/result?json", data=data)
        self.assertEqual(response.status_code, 200)

    @patch("ibkr_report.tools._MAXCACHE", 5)
    def test_cache_pruning(self):
        Cache.set("my_key", "my_value")
        self.assertEqual(Cache.get("my_key"), "my_value")

        value = "x {}"
        for key in range(5):
            Cache.set(key, value.format(key))
        for key in range(5):
            self.assertEqual(Cache.get(key), value.format(key))

        self.assertIsNone(Cache.get("my_key"))

    def test_str_enum(self):
        self.assertEqual(FieldValue.TRADES, "Trades")

    def test_strtobool(self):
        self.assertTrue(_strtobool("TRUE"))
        self.assertFalse(_strtobool("FALSE"))
        with self.assertRaises(ValueError):
            _strtobool("not_a_bool")

    def test_sri(self):
        with NamedTemporaryFile() as tempf:
            val1 = calculate_sri_on_file(tempf.name)
            tempf.write(TEST_STRING.encode("utf-8"))
            tempf.flush()
            val2 = calculate_sri_on_file(tempf.name)
        self.assertEqual(val1, EMPTY_FILE_SRI)
        self.assertEqual(val2, TEST_STRING_SRI)


if __name__ == "__main__":
    unittest.main()



"""
Total selling prices, total capital gains, and total capital losses calculated
from the CSV files.
"""

import csv
from codecs import iterdecode
from decimal import Decimal
from typing import Iterable, List, Optional, Tuple

from ibkr_report.definitions import (
    CURRENCY,
    FIELD_COUNT,
    OFFSET_DICT,
    USE_DEEMED_ACQUISITION_COST,
    AssetCategory,
    DataDiscriminator,
    Field,
    FieldValue,
    ReportOptions,
    TradeDetails,
)
from ibkr_report.exchangerates import ExchangeRates
from ibkr_report.trade import Trade


class Report:
    """Total selling prices, total capital gains, and total capital losses
    calculated from the CSV files.

    When calculating the amount of profit or loss, you can deduct the deemed
    acquisition cost from the selling price of the shares, instead of deducting
    the purchase price of the shares as well as the expenses incurred in
    making a profit.

    Args:
        file (Iterable[bytes]): The input file in CSV format.
        report_currency (str): The currency used in the output.
        use_deemed_acquisition_cost (bool): Whether to use the deemed acquisition cost
                                            if it benefits you.

    Attributes:
        prices (Decimal): Total selling prices.
        gains (Decimal): Total capital gains.
        losses (Decimal): Total capital losses.
        details(List[TradeDetails]): Details from trades such as dates and quantities.
        options (ReportOptions): Report currency, whether to use the deemed
                                 acquisition cost.
        rates (ExchangeRates): Euro foreign exchange rates.
    """

    prices: Decimal = Decimal(0)
    gains: Decimal = Decimal(0)
    losses: Decimal = Decimal(0)
    details: List[TradeDetails]
    options: ReportOptions
    rates: ExchangeRates
    _trade: Optional[Trade] = None

    def __init__(
        self,
        file: Iterable[bytes] = None,
        report_currency: str = CURRENCY,
        use_deemed_acquisition_cost: bool = USE_DEEMED_ACQUISITION_COST,
    ) -> None:
        self.details = []
        self.options = ReportOptions(
            report_currency=report_currency.upper(),
            deemed_acquisition_cost=use_deemed_acquisition_cost,
            offset=0,
        )
        self.rates = ExchangeRates()
        if file:
            self.add_trades(file)

    def add_trades(self, file: Iterable[bytes]) -> None:
        """Adds trades from a CSV formatted report file."""
        try:
            for items_list in csv.reader(iterdecode(file, "utf-8")):
                items = tuple(items_list)
                self._handle_one_line(items)
        except UnicodeDecodeError as err:
            raise ValueError("Input data not in UTF-8 text format.") from err

    def is_stock_or_options_trade(self, items: Tuple[str, ...]) -> bool:
        """Checks whether the current row is part of a trade or not."""
        if (
            len(items) == FIELD_COUNT + self.options.offset
            and items[Field.TRADES] == FieldValue.TRADES
            and items[Field.HEADER] == FieldValue.HEADER
            and items[Field.DATA_DISCRIMINATOR]
            in (DataDiscriminator.TRADE, DataDiscriminator.CLOSED_LOT)
            and items[Field.ASSET_CATEGORY]
            in (AssetCategory.STOCKS, AssetCategory.OPTIONS)
        ):
            return True
        return False

    def _handle_one_line(self, items: Tuple[str, ...]) -> None:
        offset = OFFSET_DICT.get(items)
        if offset is not None:
            self.options.offset = offset
            self._trade = None
            return
        if self.is_stock_or_options_trade(items):
            self._handle_trade(items)

    def _handle_trade(self, items: Tuple[str, ...]) -> None:
        """Parses prices, gains, and losses from trades."""
        if items[Field.DATA_DISCRIMINATOR] == DataDiscriminator.TRADE:
            self._trade = Trade(items, self.options, self.rates)
            self.prices += self._trade.total_selling_price
        if items[Field.DATA_DISCRIMINATOR] == DataDiscriminator.CLOSED_LOT:
            if not self._trade:
                raise ValueError("Tried to close a lot without trades.")
            details = self._trade.details_from_closed_lot(items)
            if details.realized > 0:
                self.gains += details.realized
            else:
                self.losses -= details.realized
            self.details.append(details)



"""Euro foreign exchange rates from European Central Bank"""

import csv
import logging
import re
from codecs import iterdecode
from datetime import timedelta
from decimal import Decimal
from io import BytesIO
from typing import Iterable
from urllib.error import HTTPError
from urllib.request import urlopen
from zipfile import BadZipFile, ZipFile

from ibkr_report.definitions import (
    DATE_FORMAT,
    EXCHANGE_RATES_URL,
    MAX_BACKTRACK_DAYS,
    MAX_HTTP_RETRIES,
    CurrencyDict,
    StorageType,
)
from ibkr_report.tools import get_date, is_number
from ibkr_report.storage import get_storage

log = logging.getLogger(__name__)


class ExchangeRates:
    """Euro foreign exchange rates"""

    rates: CurrencyDict

    def __init__(
        self, url: str = None, storage_type: StorageType = None, **kwargs
    ) -> None:
        """Tries to fetch a previously built exchange rate dictionary from a
        Storage backend. If that's not available, downloads the official
        exchange rates from European Central Bank and builds a new dictionary
        from it.
        """
        url = url or EXCHANGE_RATES_URL
        storage = get_storage(storage_type=storage_type)
        self.storage = storage(**kwargs)
        self.rates = self.storage.load()
        if not self.rates:
            self.download_official_rates(url)
            self.storage.save(content=self.rates)

    def add_to_exchange_rates(self, rates_file: Iterable[bytes]) -> None:
        """Builds the dictionary for the exchange rates from the downloaded CSV file
        and adds it to the dictionary.

        {"2015-01-20": {"USD": "1.1579", ...}, ...}
        """
        rates: CurrencyDict = {}
        currencies = []
        for items in csv.reader(iterdecode(rates_file, "utf-8")):
            if items[0] == "Date":
                # The first row should be "Date,USD,JPY,..."
                currencies = items[1:]
            if currencies and re.match(r"^\d\d\d\d-\d\d-\d\d$", items[0]):
                # And the following rows like "2015-01-20,1.1579,137.37,..."
                date_rates = {
                    cur: val
                    for cur, val in zip(currencies, items[1:])
                    if is_number(val)
                }
                if date_rates:
                    rates[items[0]] = date_rates
        log.debug("Adding currency data from %d rows.", len(rates))
        self.rates = {**self.rates, **rates}
        # TODO: Python3.9+ "self.rates |= rates"  # pylint: disable=fixme

    def download_official_rates(self, url: str) -> None:
        """Downloads the official currency exchange rates from European Central Bank
        and builds a new exchange rate dictionary from it.
        """
        retries = 0
        while True:
            if retries > MAX_HTTP_RETRIES:
                raise ValueError(
                    "Maximum number of retries exceeded. "
                    "Could not retrieve currency exchange rates."
                )
            try:
                with urlopen(url) as response:
                    bytes_io = BytesIO(response.read())
                break
            except HTTPError as err:
                # Return code error (e.g. 404, 501, ...)
                error_msg = "HTTP Error while retrieving rates: %d %s"
                log.warning(error_msg, err.code, err.reason)
                retries += 1
        log.debug("Successfully downloaded the latest exchange rates: %s", url)
        self.unzip_and_add(bytes_io)

    def unzip_and_add(self, bytes_io: BytesIO) -> None:
        """Unzips the data and passes it to `add_to_exchange_rates`."""
        try:
            with ZipFile(bytes_io) as rates_zip:
                for filename in rates_zip.namelist():
                    with rates_zip.open(filename) as rates_file:
                        self.add_to_exchange_rates(rates_file)
        except BadZipFile:
            bytes_io.seek(0)
            self.add_to_exchange_rates(bytes_io)
        log.debug("Parsed exchange rates from the retrieved data.")

    def get_rate(self, currency_from: str, currency_to: str, date_str: str) -> Decimal:
        """Exchange rate between two currencies on a given day."""
        if currency_from == currency_to:
            return Decimal(1)

        original_date = search_date = get_date(date_str)
        while original_date - search_date <= timedelta(MAX_BACKTRACK_DAYS):
            date_rates = self.rates.get(search_date.strftime(DATE_FORMAT), {})
            from_rate = "1" if currency_from == "EUR" else date_rates.get(currency_from)
            to_rate = "1" if currency_to == "EUR" else date_rates.get(currency_to)
            if from_rate is not None and to_rate is not None:
                return Decimal(to_rate) / Decimal(from_rate)
            search_date -= timedelta(1)
        raise ValueError(
            f"Currencies {currency_from} and {currency_to} not found near "
            f"date {original_date} - search ended before {search_date}"
        )



"""Storage to save and load exchange rates."""

import importlib
import json
import logging
from abc import ABC, abstractmethod
from datetime import datetime
from lzma import compress, decompress
from pathlib import Path
from typing import Optional, Type

from ibkr_report.definitions import (
    BUCKET_ID,
    DATE_FORMAT,
    SAVED_RATES_FILE,
    STORAGE_DIR,
    STORAGE_TYPE,
    CurrencyDict,
    StorageType,
)
from ibkr_report.tools import Cache

log = logging.getLogger(__name__)


class Storage(ABC):
    """Storage class to save exchange rates."""

    cache: bool = True

    def save(self, content: CurrencyDict, filename: str = None) -> None:
        """Save CurrencyDict to storage."""
        filename = filename or self.get_filename()
        log.debug("Save to '%s' using %s backend.", filename, self.name)
        if self.cache:
            Cache.set(filename, content)
        self._save(content, filename)

    def load(self, filename: str = None) -> CurrencyDict:
        """Load CurrencyDict from storage."""
        filename = filename or self.get_filename()
        log.debug("Load '%s' using %s backend.", filename, self.name)
        if self.cache:
            content = Cache.get(filename)
            if content:
                log.debug("Cache hit: %s", filename)
                return content
            log.debug("Cache miss: %s", filename)
        content = self._load(filename)
        if content and self.cache:
            Cache.set(filename, content)
        return content

    @abstractmethod
    def _save(self, content: CurrencyDict, filename: str) -> None:
        """Backend specific `save` implementation."""

    @abstractmethod
    def _load(self, filename: str) -> CurrencyDict:
        """Backend specific `load` implementation."""

    @property
    def name(self) -> str:
        """Class name"""
        return self.__class__.__name__

    @staticmethod
    def get_filename(identifier: str = None) -> str:
        """Generate a filename based on the current date.

        Optionally you can give your own identifier that distinguishes files.
        """
        if identifier is None:
            identifier = datetime.now().strftime(DATE_FORMAT)
        return SAVED_RATES_FILE.format(identifier)

    @staticmethod
    def encode(data: CurrencyDict) -> bytes:
        """Encode dictionary so it can be saved."""
        return compress(json.dumps(data).encode("utf-8"))

    @staticmethod
    def decode(data: bytes) -> CurrencyDict:
        """Decode saved dictionary."""
        return json.loads(decompress(data).decode("utf-8"))


class StorageDisabled(Storage):
    """Storage backend that doesn't do anything."""

    def _save(self, content: CurrencyDict, filename: str) -> None:
        pass

    def _load(self, filename: str) -> CurrencyDict:
        return {}


class AmazonS3(Storage):
    """Amazon S3 backend"""

    def __init__(self, bucket_id: str = None) -> None:
        boto3 = importlib.import_module("boto3")

        log.debug("Using %s backend.", self.name)
        self.bucket_id = bucket_id or BUCKET_ID
        self.aws_s3 = boto3.resource("s3")  # type: ignore
        bucket = self.aws_s3.Bucket(self.bucket_id)
        bucket.create()

    def _save(self, content: CurrencyDict, filename: str) -> None:
        obj = self.aws_s3.Object(self.bucket_id, filename)
        obj.put(Body=self.encode(content))

    def _load(self, filename: str) -> CurrencyDict:
        obj = self.aws_s3.Object(self.bucket_id, filename)
        try:
            return self.decode(obj.get()["Body"].read())
        except self.aws_s3.meta.client.exceptions.NoSuchKey:
            return {}


class GoogleCloudStorage(Storage):
    """Google Cloud Storage backend"""

    def __init__(self, bucket_id: str = None) -> None:
        storage = importlib.import_module("google.cloud.storage")
        exceptions = importlib.import_module("google.cloud.exceptions")

        log.debug("Using %s backend.", self.name)
        self.bucket_id = bucket_id or BUCKET_ID
        client = storage.Client()  # type: ignore
        try:
            self.bucket = client.get_bucket(self.bucket_id)
        except exceptions.NotFound:  # type: ignore
            self.bucket = client.create_bucket(self.bucket_id)

    def _save(self, content: CurrencyDict, filename: str) -> None:
        blob = self.bucket.blob(filename)
        blob.upload_from_string(self.encode(content))

    def _load(self, filename: str) -> CurrencyDict:
        blob = self.bucket.get_blob(filename)
        if blob:
            return self.decode(blob.download_as_bytes())
        return {}


class LocalStorage(Storage):
    """Local storage backend."""

    def __init__(self, storage_dir: str) -> None:
        log.debug("Using %s backend.", self.name)
        self.storage_dir = Path(storage_dir or STORAGE_DIR)
        self.storage_dir.mkdir(parents=True, exist_ok=True)

    def _save(self, content: CurrencyDict, filename: str) -> None:
        with open(self.storage_dir.joinpath(filename), "wb") as file:
            file.write(self.encode(content))

    def _load(self, filename: str) -> CurrencyDict:
        try:
            with open(self.storage_dir.joinpath(filename), "rb") as file:
                return self.decode(file.read())
        except FileNotFoundError:
            return {}


def get_storage(storage_type: StorageType = None) -> Type[Storage]:
    """Returns a storage backend."""
    if storage_type is None:
        storage_type = StorageType(STORAGE_TYPE)

    cloud_storage = get_cloud_storage(storage_type)
    if cloud_storage:
        return cloud_storage

    if storage_type is StorageType.LOCAL:
        return LocalStorage
    if storage_type is StorageType.DISABLED:
        return StorageDisabled

    raise NotImplementedError(f"Not implemented: {storage_type!r}")


def get_cloud_storage(storage_type: StorageType = None) -> Optional[Type[Storage]]:
    """Cloud storages with storage buckets."""
    if storage_type is StorageType.AWS:
        return AmazonS3
    if storage_type is StorageType.GCP:
        return GoogleCloudStorage
    # Past the cloud storage options, fail if BUCKET_ID is set
    if BUCKET_ID:
        raise ValueError(
            f"[BUCKET_ID] set as {BUCKET_ID!r}, but [STORAGE_TYPE] is set as {STORAGE_TYPE!r}."
            " With a bucket [STORAGE_TYPE] needs to be set as [AWS|GCP]."
        )
    return None



""" IBKR Report Parser

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program. If not, see <https://www.gnu.org/licenses/>.
"""
from ibkr_report import create_app, definitions


def main() -> None:
    """This is used when running the module locally."""
    app = create_app()
    app.run(host="127.0.0.1", port=8080, debug=definitions.DEBUG)


if __name__ == "__main__":
    main()



"""Constants and other definitions"""

import os
from dataclasses import dataclass
from decimal import Decimal
from enum import Enum, IntEnum, unique
from typing import Dict


def _strtobool(val: str) -> bool:
    """Convert a string representation of truth to True or False.

    True values are 'y', 'yes', 't', 'true', 'on', and '1'; false values
    are 'n', 'no', 'f', 'false', 'off', and '0'. Raises ValueError if
    'val' is anything else.
    """
    val = val.lower()
    if val in ("y", "yes", "t", "true", "on", "1"):
        return True
    if val in ("n", "no", "f", "false", "off", "0"):
        return False
    raise ValueError(f"invalid truth value {val!r}")


TITLE = os.getenv("TITLE", "IBKR Report Parser")
BUCKET_NAME = os.getenv("BUCKET_NAME", None)
BUCKET_ID = os.getenv("BUCKET_ID", BUCKET_NAME)
CURRENCY = os.getenv("CURRENCY", "EUR")
USE_DEEMED_ACQUISITION_COST = _strtobool(
    os.getenv("USE_DEEMED_ACQUISITION_COST", "TRUE")
)
STORAGE_TYPE = os.getenv("STORAGE_TYPE", "disabled").lower()
STORAGE_DIR = os.getenv("STORAGE_DIR", ".ibkr_storage")
DEBUG = _strtobool(os.getenv("DEBUG", "FALSE"))
_DEFAULT_LOGGING = "DEBUG" if DEBUG else "INFO"
LOGGING_LEVEL = os.getenv("LOGGING_LEVEL", _DEFAULT_LOGGING)
_DEFAULT_URL = "https://www.ecb.europa.eu/stats/eurofxref/eurofxref-hist.zip"
EXCHANGE_RATES_URL = os.getenv("EXCHANGE_RATES_URL", _DEFAULT_URL)

MAX_BACKTRACK_DAYS = 7
MAX_HTTP_RETRIES = 5
SAVED_RATES_FILE = "official_ecb_exchange_rates-{0}.json.xz"

_SINGLE_ACCOUNT = (
    "Trades,Header,DataDiscriminator,Asset Category,Currency,Symbol,Date/Time,Exchange,"
    "Quantity,T. Price,Proceeds,Comm/Fee,Basis,Realized P/L,Code"
).split(",")
_MULTI_ACCOUNT = (
    "Trades,Header,DataDiscriminator,Asset Category,Currency,Account,Symbol,Date/Time,Exchange,"
    "Quantity,T. Price,Proceeds,Comm/Fee,Basis,Realized P/L,Code"
).split(",")
OFFSET_DICT = {
    tuple(_SINGLE_ACCOUNT): 0,
    tuple(_MULTI_ACCOUNT): len(_MULTI_ACCOUNT) - len(_SINGLE_ACCOUNT),
}
FIELD_COUNT = len(_SINGLE_ACCOUNT)
DATE_FORMAT = "%Y-%m-%d"
TIME_FORMAT = " %H:%M:%S"
DATE_STR_FORMATS = (
    DATE_FORMAT + "," + TIME_FORMAT,
    DATE_FORMAT + TIME_FORMAT,
    DATE_FORMAT,
)

CurrencyDict = Dict[str, Dict[str, str]]


class StrEnum(str, Enum):
    """TODO: StrEnum available in Python 3.11+"""

    ...


@unique
class Field(IntEnum):
    """CSV indices."""

    TRADES = 0
    HEADER = 1
    DATA_DISCRIMINATOR = 2
    ASSET_CATEGORY = 3
    CURRENCY = 4
    SYMBOL = 5
    DATE_TIME = 6
    EXCHANGE = 7
    QUANTITY = 8
    TRANSACTION_PRICE = 9
    PROCEEDS = 10
    COMMISSION_AND_FEES = 11
    BASIS = 12
    REALIZED_PL = 13
    CODE = 14


@unique
class DataDiscriminator(StrEnum):
    """CSV DataDiscriminator values"""

    TRADE = "Trade"
    CLOSED_LOT = "ClosedLot"


@unique
class AssetCategory(StrEnum):
    """CSV Asset Category values"""

    STOCKS = "Stocks"
    OPTIONS = "Equity and Index Options"


@unique
class FieldValue(StrEnum):
    """Other possible values in a CSV file"""

    TRADES = "Trades"
    HEADER = "Data"


@unique
class StorageType(StrEnum):
    """Storage type"""

    DISABLED = "disabled"
    AWS = "aws"
    GCP = "gcp"
    LOCAL = "local"


@dataclass
class ReportOptions:
    """Report options"""

    report_currency: str
    deemed_acquisition_cost: bool
    offset: int


@dataclass
class RowData:
    """Extracted data from a CSV file"""

    symbol: str
    date_str: str
    rate: Decimal
    price_per_share: Decimal
    quantity: Decimal


@dataclass
class TradeDetails:
    """Extracted and calculated data from a trade"""

    symbol: str
    quantity: Decimal
    buy_date: str
    sell_date: str
    price: Decimal
    realized: Decimal



"""Flask Blueprints for the website."""

from typing import Dict, Union
from flask import Blueprint, Response, abort, make_response, render_template, request

from ibkr_report.definitions import TITLE
from ibkr_report.report import Report
from ibkr_report.tools import set_logging, _sri


bp = Blueprint("website", __name__)


@bp.route("/", methods=["GET"])
def index() -> Response:
    """Main page"""
    resp = make_response(render_template("index.html", title=TITLE, sri=_sri()))
    resp.cache_control.max_age = 600
    return resp


@bp.route("/result", methods=["POST"])
def result():
    """Parse results and pass them to show_results."""
    set_logging()
    try:
        report = Report(request.files.get("file"))
    except ValueError as err:
        abort(400, description=err)
    json_format = request.args.get("json") is not None
    return show_results(report=report, json_format=json_format)


def show_results(
    report: Report, json_format: bool = False
) -> Union[str, Dict[str, object]]:
    """Show the results either in JSON or HTML format."""
    if json_format:
        return {
            "prices": float(round(report.prices, 2)),
            "gains": float(round(report.gains, 2)),
            "losses": float(round(report.losses, 2)),
            "details": report.details,
        }
    return render_template(
        "result.html",
        title=TITLE,
        prices=report.prices,
        gains=report.gains,
        losses=report.losses,
        details=report.details,
        sri=_sri(),
    )


@bp.errorhandler(400)
def bad_request(err):
    """Error response for bad requests."""
    if request.args.get("json") is not None:
        return {"error": str(err)}, 400
    return (
        render_template(
            "error.html",
            title=TITLE,
            message=str(err),
            sri=_sri(),
        ),
        400,
    )



import os
from setuptools import setup, find_packages  # type: ignore

NAME = "ibkr-report-parser"
PACKAGES = find_packages()

DESCRIPTION = "Interactive Brokers (IBKR) Report Parser for MyTax (vero.fi)"
URL = "https://github.com/oittaa/ibkr-report-parser"
LONG_DESCRIPTION = open(os.path.join(os.path.dirname(__file__), "README.md")).read()

AUTHOR = "Oittaa"
AUTHOR_EMAIL = ""
GITHUB_REF = os.environ.get("GITHUB_REF")
PREFIX = "refs/tags/"

if GITHUB_REF and GITHUB_REF.startswith(PREFIX):
    prefix_len = len(PREFIX)
    VERSION = GITHUB_REF[prefix_len:]
else:
    VERSION = "0.0.0.dev0"

setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url=URL,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    packages=PACKAGES,
    include_package_data=True,
    zip_safe=False,
    keywords=[
        "Interactive Brokers",
        "IBKR",
        "OmaVero",
        "MyTax",
        "vero.fi",
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Other Audience",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    entry_points={
        "console_scripts": [
            "ibkr-report-parser=ibkr_report.__main__:main",
        ],
    },
    setup_requires=[
        "wheel",
    ],
    install_requires=["flask==2.0.2"],
    extras_require={
        "aws": ["boto3==1.20.39"],
        "docker": ["gunicorn==20.1.0"],
        "gcp": ["google-cloud-storage==2.1.0"],
    },
    python_requires=">=3.7",
)



import os
import unittest
from decimal import Decimal
from unittest.mock import patch

from ibkr_report import ExchangeRates
from ibkr_report.tools import Cache

THIS_PATH = os.path.abspath(os.getcwd())
TEST_URL = f"file://{THIS_PATH}/tests/test-data/eurofxref-hist.zip"
TEST_BROKEN_URL = f"file://{THIS_PATH}/tests/test-data/eurofxref-broken.csv"


class ExchangeTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.rates = ExchangeRates(TEST_URL)

    def setUp(self):
        Cache.clear()

    def test_self(self):
        same_same = self.rates.get_rate("USD", "USD", "2015-12-01")
        self.assertEqual(same_same, 1)

    def test_sek_usd(self):
        sek_usd = self.rates.get_rate("SEK", "USD", "2015-12-01")
        self.assertLess(sek_usd, Decimal("0.2"))

    def test_back_and_forth(self):
        sek_usd = self.rates.get_rate("SEK", "USD", "2015-12-01")
        usd_sek = self.rates.get_rate("USD", "SEK", "2015-12-01")
        should_be_one = sek_usd * usd_sek
        self.assertEqual(should_be_one, 1)
        self.assertNotEqual(sek_usd, 1)

    def test_from_euro(self):
        eur_usd = self.rates.get_rate("EUR", "USD", "2015-12-01")
        self.assertGreater(eur_usd, 1)

    def test_to_euro(self):
        nok_eur = self.rates.get_rate("NOK", "EUR", "2010-01-01")
        self.assertLess(nok_eur, Decimal("0.2"))

    @patch("ibkr_report.exchangerates.MAX_BACKTRACK_DAYS", 0)
    def test_no_backtrack(self):
        with self.assertRaises(ValueError):
            self.rates.get_rate("NOK", "EUR", "2010-01-01")

    def test_currency_does_not_exist(self):
        with self.assertRaises(ValueError):
            self.rates.get_rate("KEKW", "USD", "2015-12-01")

    def test_far_in_the_future(self):
        with self.assertRaises(ValueError):
            self.rates.get_rate("USD", "CAD", "2500-01-01")

    def test_broken_input(self):
        with open("tests/test-data/eurofxref-broken.csv", "rb") as file:
            self.rates.add_to_exchange_rates(file)
        eur_usd = self.rates.get_rate("EUR", "USD", "2021-10-26")
        self.assertEqual(eur_usd, Decimal("1.1618"))
        eur_usd = self.rates.get_rate("EUR", "USD", "2021-10-25")
        self.assertEqual(eur_usd, Decimal("1.1603"))

    def test_download_without_zip(self):
        rates = ExchangeRates(TEST_BROKEN_URL)
        eur_usd = rates.get_rate("EUR", "USD", "2021-10-26")
        self.assertEqual(eur_usd, Decimal("1.1618"))
        with self.assertRaises(ValueError):
            eur_usd = rates.get_rate("EUR", "USD", "2021-10-25")


if __name__ == "__main__":
    unittest.main()
